"""
Handler Module with all the individual handlers for the jupyterlab-gitextension.
"""
import json


from notebook.utils import url_path_join as ujoin
from notebook.base.handlers import APIHandler


class Git_handler(APIHandler):
    """
    Git Parent Handler.
    """

    @property
    def git(self):
        return self.settings["git"]


class Git_all_history_handler(Git_handler):
    """
    Parent handler for all four history/status git commands:
    1. git show_top_level
    2. git branch
    3. git log
    4. git status
    Called on refresh of extension's widget
    """

    def post(self):
        """
        POST request handler, calls individual handlers for 
        'git show_top_level', 'git branch', 'git log', and 'git status'
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        show_top_level = self.git.show_top_level(current_path)
        if show_top_level["code"] != 0:
            self.finish(json.dumps(show_top_level))
        else:
            branch = self.git.branch(current_path)
            log = self.git.log(current_path)
            status = self.git.status(current_path)

            result = {
                "code": show_top_level["code"],
                "data": {
                    "show_top_level": show_top_level,
                    "branch": branch,
                    "log": log,
                    "status": status,
                },
            }
            self.finish(json.dumps(result))


class Git_show_top_level_handler(Git_handler):
    """
    Handler for 'git rev-parse --show-toplevel'. 
    Displays the git root directory inside a repository.
    """

    def post(self):
        """
        POST request handler, displays the git root directory inside a repository.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        result = self.git.show_top_level(current_path)
        self.finish(json.dumps(result))


class Git_show_prefix_handler(Git_handler):
    """
    Handler for 'git rev-parse --show-prefix'. 
    Displays the prefix path of a directory in a repository, 
    with respect to the root directory.
    """

    def post(self):
        """
        POST request handler, displays the prefix path of a directory in a repository, 
        with respect to the root directory.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        result = self.git.show_prefix(current_path)
        self.finish(json.dumps(result))


class Git_status_handler(Git_handler):
    """
    Handler for 'git status --porcelain', fetches the git status.
    """

    def get(self):
        """
        GET request handler, shows file status, used in refresh method.
        """
        self.finish(
            json.dumps(
                {"add_all": "check", "filename": "filename", "top_repo_path": "path"}
            )
        )

    def post(self):
        """
        POST request handler, fetches the git status.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        result = self.git.status(current_path)
        self.finish(json.dumps(result))


class Git_log_handler(Git_handler):
    """
    Handler for 'git log --pretty=format:%H-%an-%ar-%s'. 
    Fetches Commit SHA, Author Name, Commit Date & Commit Message.
    """

    def post(self):
        """
        POST request handler, 
        fetches Commit SHA, Author Name, Commit Date & Commit Message.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        result = self.git.log(current_path)
        self.finish(json.dumps(result))


class Git_detailed_log_handler(Git_handler):
    """
    Handler for 'git log -1 --stat --numstat --oneline' command. 
    Fetches file names of committed files, Number of insertions &
    deletions in that commit.
    """

    def post(self):
        """
        POST request handler, fetches file names of committed files, Number of
        insertions & deletions in that commit.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        selected_hash = my_data["selected_hash"]
        current_path = my_data["current_path"]
        result = self.git.detailed_log(selected_hash, current_path)
        self.finish(json.dumps(result))


class Git_diff_handler(Git_handler):
    """
    Handler for 'git diff --numstat'. Fetches changes between commits & working tree.
    """

    def post(self):
        """
        POST request handler, fetches differences between commits & current working
        tree.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        top_repo_path = my_data["top_repo_path"]
        my_output = self.git.diff(top_repo_path)
        self.finish(my_output)
        print("GIT DIFF")
        print(my_output)


class Git_branch_handler(Git_handler):
    """
    Handler for 'git branch -a'. Fetches list of all branches in current repository
    """

    def post(self):
        """
        POST request handler, fetches all branches in current repository.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        result = self.git.branch(current_path)
        self.finish(json.dumps(result))


class Git_add_handler(Git_handler):
    """
    Handler for git add <filename>'. 
    Adds one or all files into to the staging area.
    """

    def get(self):
        """
        GET request handler, adds files in the staging area.
        """
        self.finish(
            json.dumps(
                {"add_all": "check", "filename": "filename", "top_repo_path": "path"}
            )
        )

    def post(self):
        """
        POST request handler, adds one or all files into the staging area.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        top_repo_path = my_data["top_repo_path"]
        if my_data["add_all"]:
            my_output = self.git.add_all(top_repo_path)
        else:
            filename = my_data["filename"]
            my_output = self.git.add(filename, top_repo_path)
        self.finish(my_output)


class Git_reset_handler(Git_handler):
    """
    Handler for 'git reset <filename>'. 
    Moves one or all files from the staged to the unstaged area.
    """

    def post(self):
        """
        POST request handler, 
        moves one or all files from the staged to the unstaged area.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        top_repo_path = my_data["top_repo_path"]
        if my_data["reset_all"]:
            my_output = self.git.reset_all(top_repo_path)
        else:
            filename = my_data["filename"]
            my_output = self.git.reset(filename, top_repo_path)
        self.finish(my_output)


class Git_checkout_handler(Git_handler):
    """
    Handler for 'git checkout <branchname>'. Changes the current working branch.
    """

    def post(self):
        """
        POST request handler, changes between branches.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        top_repo_path = my_data["top_repo_path"]
        if my_data["checkout_branch"]:
            if my_data["new_check"]:
                print("to create a new branch")
                my_output = self.git.checkout_new_branch(
                    my_data["branchname"], top_repo_path
                )
            else:
                print("switch to an old branch")
                my_output = self.git.checkout_branch(
                    my_data["branchname"], top_repo_path
                )
        elif my_data["checkout_all"]:
            my_output = self.git.checkout_all(top_repo_path)
        else:
            my_output = self.git.checkout(my_data["filename"], top_repo_path)
        self.finish(my_output)


class Git_commit_handler(Git_handler):
    """
    Handler for 'git commit -m <message>'. Commits files.
    """

    def post(self):
        """
        POST request handler, commits files.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        top_repo_path = my_data["top_repo_path"]
        commit_msg = my_data["commit_msg"]
        my_output = self.git.commit(commit_msg, top_repo_path)
        self.finish(my_output)


class Git_pull_handler(Git_handler):
    """
    Handler for 'git pull <first-branch> <second-branch>'. Pulls files from a remote branch.
    """

    def post(self):
        """
        POST request handler, pulls files from a remote branch to your current branch.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        origin = my_data["origin"]
        master = my_data["master"]
        curr_fb_path = my_data["curr_fb_path"]
        my_output = self.git.pull(origin, master, curr_fb_path)
        self.finish(my_output)
        print("You Pulled")


class Git_push_handler(Git_handler):
    """
    Handler for 'git push <first-branch> <second-branch>. 
    Pushes committed files to a remote branch.
    """

    def post(self):
        """
        POST request handler, 
        pushes comitted files from your current branch to a remote branch
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        origin = my_data["origin"]
        master = my_data["master"]
        curr_fb_path = my_data["curr_fb_path"]
        my_output = self.git.push(origin, master, curr_fb_path)
        self.finish(my_output)
        print("You Pushed")


class Git_init_handler(Git_handler):
    """
    Handler for 'git init'. Initializes a repository.
    """

    def post(self):
        """
        POST request handler, initializes a repository.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        current_path = my_data["current_path"]
        my_output = self.git.init(current_path)
        self.finish(my_output)


class Git_add_all_untracked_handler(Git_handler):
    """
    Handler for 'echo "a\n*\nq\n" | git add -i'. Adds ONLY all untracked files.
    """

    def post(self):
        """
        POST request handler, adds all the untracked files.
        """
        my_data = json.loads(self.request.body.decode("utf-8"))
        top_repo_path = my_data["top_repo_path"]
        my_output = self.git.add_all_untracked(top_repo_path)
        print(my_output)
        self.finish(my_output)


def setup_handlers(web_app):
    """
    Setups all of the git command handlers.
    Every handler is defined here, to be used in git.py file.
    """

    git_handlers = [
        ("/git/show_top_level", Git_show_top_level_handler),
        ("/git/show_prefix", Git_show_prefix_handler),
        ("/git/add", Git_add_handler),
        ("/git/status", Git_status_handler),
        ("/git/branch", Git_branch_handler),
        ("/git/reset", Git_reset_handler),
        ("/git/checkout", Git_checkout_handler),
        ("/git/commit", Git_commit_handler),
        ("/git/pull", Git_pull_handler),
        ("/git/push", Git_push_handler),
        ("/git/diff", Git_diff_handler),
        ("/git/log", Git_log_handler),
        ("/git/detailed_log", Git_detailed_log_handler),
        ("/git/init", Git_init_handler),
        ("/git/all_history", Git_all_history_handler),
        ("/git/add_all_untracked", Git_add_all_untracked_handler),
    ]

    # add the baseurl to our paths
    base_url = web_app.settings["base_url"]
    git_handlers = [(ujoin(base_url, x[0]), x[1]) for x in git_handlers]
    print("base_url: {}".format(base_url))
    print(git_handlers)

    web_app.add_handlers(".*", git_handlers)



"""
Module for executing git commands, their results
are sent back to the caller.
"""
import os
import subprocess
from subprocess import Popen, PIPE


class Git:
    """
    A single parent Git Class which has all the individual git methods in it.
    """

    def status(self, current_path):
        """
        Execute git status command & send back the result.
        """
        p = Popen(
            ["git", "status", "--porcelain"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            for line in line_array:
                to1 = None
                from_path = line[3:]
                if line[0] == "R":
                    to0 = line[3:].split(" -> ")
                    to1 = to0[len(to0) - 1]
                else:
                    to1 = line[3:]
                if to1.startswith('"'):
                    to1 = to1[1:]
                if to1.endswith('"'):
                    to1 = to1[:-1]
                result.append(
                    {"x": line[0], "y": line[1], "to": to1, "from": from_path}
                )
            return {"code": p.returncode, "files": result}
        else:
            return {
                "code": p.returncode,
                "command": "git status --porcelain",
                "message": my_error.decode("utf-8"),
            }

    def log(self, current_path):
        """
        Execute git log command & send back the result.
        """
        p = Popen(
            ["git", "log", "--pretty=format:%H%n%an%n%ar%n%s", "-10"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            i = 0
            PREVIOUS_COMMIT_OFFSET = 4
            while i < len(line_array):
                if i + PREVIOUS_COMMIT_OFFSET < len(line_array):
                    result.append(
                        {
                            "commit": line_array[i],
                            "author": line_array[i + 1],
                            "date": line_array[i + 2],
                            "commit_msg": line_array[i + 3],
                            "pre_commit": line_array[i + PREVIOUS_COMMIT_OFFSET],
                        }
                    )
                else:
                    result.append(
                        {
                            "commit": line_array[i],
                            "author": line_array[i + 1],
                            "date": line_array[i + 2],
                            "commit_msg": line_array[i + 3],
                            "pre_commit": "",
                        }
                    )
                i += PREVIOUS_COMMIT_OFFSET
            return {"code": p.returncode, "commits": result}
        else:
            return {"code": p.returncode, "message": my_error.decode("utf-8")}

    def detailed_log(self, selected_hash, current_path):
        """
        Execute the detailed git log command (used to get
        insertions & deletions per file) & send back the result.
        """
        p = Popen(
            ["git", "log", "-1", "--stat", "--numstat", "--oneline", selected_hash],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            note = [0] * 3
            count = 0
            temp = ""
            line_array = my_output.decode("utf-8").splitlines()
            length = len(line_array)
            INSERTION_INDEX = 0
            DELETION_INDEX = 1
            MODIFIED_FILE_PATH_INDEX = 2
            if length > 1:
                temp = line_array[length - 1]
                words = temp.split()
                for i in range(0, len(words)):
                    if words[i].isdigit():
                        note[count] = words[i]
                        count += 1
                for num in range(1, int(length / 2)):
                    line_info = line_array[num].split()
                    words = line_info[2].split("/")
                    length = len(words)
                    result.append(
                        {
                            "modified_file_path": line_info[MODIFIED_FILE_PATH_INDEX],
                            "modified_file_name": words[length - 1],
                            "insertion": line_info[INSERTION_INDEX],
                            "deletion": line_info[DELETION_INDEX],
                        }
                    )

            if note[2] == 0 and length > 1:
                if "-" in temp:
                    exchange = note[1]
                    note[1] = note[2]
                    note[2] = exchange

            return {
                "code": p.returncode,
                "modified_file_note": temp,
                "modified_files_count": note[0],
                "number_of_insertions": note[1],
                "number_of_deletions": note[2],
                "modified_files": result,
            }
        else:
            return {
                "code": p.returncode,
                "command": "git log_1",
                "message": my_error.decode("utf-8"),
            }

    def diff(self, top_repo_path):
        """
        Execute git diff command & send back the result.
        """
        p = Popen(
            ["git", "diff", "--numstat"], stdout=PIPE, stderr=PIPE, cwd=top_repo_path
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            for line in line_array:
                linesplit = line.split()
                result.append(
                    {
                        "insertions": linesplit[0],
                        "deletions": linesplit[1],
                        "filename": linesplit[2],
                    }
                )
            return {"code": p.returncode, "result": result}
        else:
            return {"code": p.returncode, "message": my_error.decode("utf-8")}

    def branch(self, current_path):
        """
        Execute git branch status command & send back the result.
        """
        p = Popen(
            ["git", "branch", "-a"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            """By comparing strings 'remotes/' to determine if a branch is
            local or remote, should have better ways
            """
            for line_full in line_array:
                line_cut = (line_full.split(" -> "),)
                tag = (None,)
                current = (False,)
                remote = (False,)
                if len(line_cut[0]) > 1:
                    tag = line_cut[0][1]
                line = (line_cut[0][0],)
                if line_full[0] == "*":
                    current = (True,)
                if (len(line_full) >= 10) and (line_full[2:10] == "remotes/"):
                    remote = (True,)
                    result.append(
                        {
                            "current": current,
                            "remote": remote,
                            "name": line[0][10:],
                            "tag": tag,
                        }
                    )
                else:
                    result.append(
                        {
                            "current": current,
                            "remote": remote,
                            "name": line_full[2:],
                            "tag": tag,
                        }
                    )
            return {"code": p.returncode, "branches": result}
        else:
            return {
                "code": p.returncode,
                "command": "git branch -a",
                "message": my_error.decode("utf-8"),
            }

    def show_top_level(self, current_path):
        """
        Execute git --show-toplevel command & send back the result.
        """
        p = Popen(
            ["git", "rev-parse", "--show-toplevel"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = {
                "code": p.returncode,
                "top_repo_path": my_output.decode("utf-8").strip("\n"),
            }
            return result
        else:
            return {
                "code": p.returncode,
                "command": "git rev-parse --show-toplevel",
                "message": my_error.decode("utf-8"),
            }

    def show_prefix(self, current_path):
        """
        Execute git --show-prefix command & send back the result.
        """
        p = Popen(
            ["git", "rev-parse", "--show-prefix"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = {
                "code": p.returncode,
                "under_repo_path": my_output.decode("utf-8").strip("\n"),
            }
            return result
        else:
            return {
                "code": p.returncode,
                "command": "git rev-parse --show-prefix",
                "message": my_error.decode("utf-8"),
            }

    def add(self, filename, top_repo_path):
        """
        Execute git add<filename> command & send back the result.
        """
        my_output = subprocess.check_output(["git", "add", filename], cwd=top_repo_path)
        return my_output

    def add_all(self, top_repo_path):
        """
        Execute git add all command & send back the result.
        """
        my_output = subprocess.check_output(["git", "add", "-u"], cwd=top_repo_path)
        return my_output

    def add_all_untracked(self, top_repo_path):
        """
        Execute git add_all_untracked command & send back the result.
        """
        e = 'echo "a\n*\nq\n" | git add -i'
        my_output = subprocess.call(e, shell=True, cwd=top_repo_path)
        return {"result": my_output}

    def reset(self, filename, top_repo_path):
        """
        Execute git reset <filename> command & send back the result.
        """
        my_output = subprocess.check_output(
            ["git", "reset", filename], cwd=top_repo_path
        )
        return my_output

    def reset_all(self, top_repo_path):
        """
        Execute git reset all command & send back the result.
        """
        my_output = subprocess.check_output(["git", "reset"], cwd=top_repo_path)
        return my_output

    def checkout_new_branch(self, branchname, current_path):
        """
        Execute git checkout <make-branch> command & send back the result.
        """
        p = Popen(
            ["git", "checkout", "-b", branchname],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return {"code": p.returncode, "message": my_output.decode("utf-8")}
        else:
            return {
                "code": p.returncode,
                "command": "git checkout " + "-b" + branchname,
                "message": my_error.decode("utf-8"),
            }

    def checkout_branch(self, branchname, current_path):
        """
        Execute git checkout <branch-name> command & send back the result.
        """
        p = Popen(
            ["git", "checkout", branchname],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + current_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return {"code": p.returncode, "message": my_output.decode("utf-8")}
        else:
            return {
                "code": p.returncode,
                "command": "git checkout " + branchname,
                "message": my_error.decode("utf-8"),
            }

    def checkout(self, filename, top_repo_path):
        """
        Execute git checkout command for the filename & send back the result.
        """
        my_output = subprocess.check_output(
            ["git", "checkout", "--", filename], cwd=top_repo_path
        )
        return my_output

    def checkout_all(self, top_repo_path):
        """
        Execute git checkout command & send back the result.
        """
        my_output = subprocess.check_output(
            ["git", "checkout", "--", "."], cwd=top_repo_path
        )
        return my_output

    def commit(self, commit_msg, top_repo_path):
        """
        Execute git commit <filename> command & send back the result.
        """
        my_output = subprocess.check_output(
            ["git", "commit", "-m", commit_msg], cwd=top_repo_path
        )
        return my_output

    def pull(self, origin, master, curr_fb_path):
        """
        Execute git pull <branch1> <branch2> command & send back the result.
        """
        p = Popen(
            ["git", "pull", origin, master, "--no-commit"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + curr_fb_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return {"code": p.returncode}
        else:
            return {
                "code": p.returncode,
                "command": "git pull " + origin + " " + master + " --no-commit",
                "message": my_error.decode("utf-8"),
            }

    def push(self, origin, master, curr_fb_path):
        """
        Execute git push <branch1> <branch2> command & send back the result.
        """
        p = Popen(
            ["git", "push", origin, master],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.getcwd() + "/" + curr_fb_path,
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return {"code": p.returncode}
        else:
            return {
                "code": p.returncode,
                "command": "git push " + origin + " " + master,
                "message": my_error.decode("utf-8"),
            }

    def init(self, current_path):
        """
        Execute git init command & send back the result.
        """
        my_output = subprocess.check_output(
            ["git", "init"], cwd=os.getcwd() + "/" + current_path
        )
        return my_output



"""
Module to initialize Server Extension & Notebook Extension
"""
from jupyterlab_git.handlers import setup_handlers
from jupyterlab_git.git import Git


def _jupyter_server_extension_paths():
    """
    Function to declare Jupyter Server Extension Paths.
    """
    return [{"module": "jupyterlab_git"}]


def _jupyter_nbextension_paths():
    """
    Function to declare Jupyter Notebook Extension Paths.
    """
    return [{"section": "notebook", "dest": "jupyterlab_git"}]


def load_jupyter_server_extension(nbapp):
    """
    Function to load Jupyter Server Extension.
    """
    git = Git()
    nbapp.web_app.settings["git"] = git
    setup_handlers(nbapp.web_app)



"""
Setup Module to setup Python Handlers (Git Handlers) for the Git Plugin.
"""
import setuptools

setuptools.setup(
    name='jupyterlab_git',
    version='0.1.1',
    author='Git Intern Team, Noah Stapp, Jenna Landy, Alena Mueller',
    description="A server extension for JupyterLab's git extension",
    packages=setuptools.find_packages(),
    install_requires=[
        'notebook',
        'psutil'
    ],
    package_data={'jupyterlab_git': ['*']},
)
